@@ -1,220 +1,14 @@
+
 import random
+
 import numpy as np
+from qsin.utils import progressbar
+from qsin.CD_dual_gap import (epoch_lasso, update_beta_lasso, dualpa, 
+                              epoch_enet , update_beta_enet , dualpa_elnet)
 
-# import matplotlib.pyplot as plt
-import copy
-from numba import njit
-
-
-# @njit
 def sparse_XB(X, B):
-    # return sparse.csr_matrix(X).multiply(sparse.csr_matrix(B))
     return np.matmul(X, B)
 
-# sparse_XB = lambda X, B: np.matmul(X, B)
-
-
-# np.random.seed(123)
-# random.seed(123)
-
-# standarize data
-def std_data(X):
-    return (X - np.mean(X, axis=0)) / np.std(X, axis=0)
-
-def mse(y, X, B):
-    return np.mean( ( y - X.dot(B) )**2 ) 
-
-
-@njit
-def msoft_threshold( delta, lam, denom):
-    if delta > lam:
-        return (delta - lam) / denom
-    elif delta < -lam:
-        return (delta + lam) / denom
-    else:
-        return 0
-
-@njit
-def update_beta_lasso(beta, lam, c1, Xj_T_y, Xj_T_Xj, X_T_X, chosen_ps):
-
-    for j in chosen_ps:
-        # j = 1
-        # A2 = np.where(beta !=  0)[0]
-        # # take out j from the array A2
-        # A2 = A2[A2 != j]
-
-        A2 = np.where(beta !=  beta[j])[0]
-
-        tmp_X_T_X = X_T_X[j,:]
-
-        delta = c1 * ( Xj_T_y[j] - np.dot(tmp_X_T_X[A2], beta[A2]) )
-        denom = c1 * Xj_T_Xj[j]
-        beta[j] = msoft_threshold(delta, lam, denom)
-
-
-@njit
-def epoch_lasso(X, beta, lam, c1, Xj_T_y, Xj_T_Xj, X_T_X, chosen_ps, s_new, diff):
-
-    for j in chosen_ps:
-        # j = 1
-        b_old = beta[j]
-        
-        A2 = np.where(beta != 0)[0]
-        # take out j from the array A2
-        A2 = A2[A2 != j]
-        tmp_X_T_X = X_T_X[j,:]
-
-        delta = c1 * ( Xj_T_y[j] - np.dot(tmp_X_T_X[A2], beta[A2]) )
-        denom = c1 * Xj_T_Xj[j]
-        beta[j] = msoft_threshold(delta, lam, denom)
-
-
-        # beautiful rank 1 sums 
-        if beta[j] != 0.0:
-            s_new += X[:,j] * beta[j]
-
-        diff_j = b_old - beta[j]
-        if diff_j != 0.0:    
-            diff += X[:,j] * diff_j
-
-
-@njit
-def epoch_enet(X, beta, c1, Xj_T_y, Xj_T_Xj, X_T_X, chosen_ps, lam_1_alpha, lam_alpha, s_new, diff):
-
-    for j in chosen_ps:
-        # j = 1
-        b_old = beta[j]
-        
-        A2 = np.where(beta != 0)[0]
-        # take out j from the array A2
-        A2 = A2[A2 != j]
-        tmp_X_T_X = X_T_X[j,:]
-
-        delta = c1 * ( Xj_T_y[j] - np.dot(tmp_X_T_X[A2], beta[A2]) )
-        denom = (c1 * Xj_T_Xj[j]) + lam_1_alpha
-
-        beta[j] = msoft_threshold(delta, lam_alpha, denom)
-
-
-        # beautiful rank 1 sums 
-        if beta[j] != 0.0:
-            s_new += X[:,j] * beta[j]
-
-        diff_j = b_old - beta[j]
-        if diff_j != 0.0:    
-            diff += X[:,j] * diff_j
-
-
-@njit
-def update_beta_enet(beta, c1, Xj_T_y, Xj_T_Xj, X_T_X, 
-                     chosen_ps, lam_1_alpha, lam_alpha):
-    for j in chosen_ps:
-        # j = 1
-        A2 = np.where(beta != 0)[0]
-        # take out j from the array A2
-        A2 = A2[A2 != j]
-
-        tmp_X_T_X = X_T_X[j,:]
-
-        delta = c1 * ( Xj_T_y[j] - np.dot(tmp_X_T_X[A2], beta[A2]) )
-        denom = (c1 * Xj_T_Xj[j]) + lam_1_alpha
-
-        beta[j] = msoft_threshold(delta, lam_alpha, denom)
-
-    # return beta
-
-
-# @jit
-def calculate_dual_gap(R, X, y, lam, beta):
-    n = len(y)
-
-    p_obj = (1/n) * np.linalg.norm(R)**2  + lam * np.linalg.norm(beta, ord=1)
-
-    theta = (2/n) * R
-    norm_inf = np.linalg.norm(X.T @ theta, ord=np.inf)
-    # max([1,2])
-    # deno = norm_inf if norm_inf > lam else lam
-    theta_ = theta/np.max([lam, norm_inf])
-
-    d_obj = (1/n)*( np.linalg.norm(y)**2 - np.linalg.norm(y - (lam*n/2)*theta_)**2 )
-
-    w_d_gap = (p_obj - d_obj)*n/np.linalg.norm(y)**2
-
-    return w_d_gap
-
-
-def dualpa_max(norm_inf, lam):
-    return lam if norm_inf < lam else norm_inf
-
-def dualpa(X, y, lam, beta, ny2):
-    R = y - X @ beta
-    
-
-    n= len(y)
-    c1 = 1/n
-
-    rt = 2*c1*R
-    norm_inf = np.linalg.norm(X.T @ rt, ord=np.inf)
-    f = dualpa_max(norm_inf, lam)
-
-    c2 = lam/f
-    
-    d_gap = (1/n)*np.dot(R, (1 + c2**2)*R - 2*c2*y) + lam*np.linalg.norm(beta, ord=1)
-    # d_gap = dualpa_gap(R, y, lam, beta, c2, n)
-
-    return d_gap*n/ny2
-
-# @jit
-def duality_gap_elnet(R, X, y, beta, lam_1_alpha, lam_alpha):
-
-    n = len(y)
-
-    R_norm_sq = np.linalg.norm(R)**2
-    b_norm_sq = np.linalg.norm(beta)**2
-
-    c1 = 1/n
-    p_obj = c1 * R_norm_sq  + lam_alpha * np.linalg.norm(beta, ord=1) + (lam_1_alpha/2)*b_norm_sq
-    rt = (2/n) * R
-
-    norm_inf = np.linalg.norm(X.T @ rt  - lam_1_alpha*beta, ord=np.inf)
-
-    # f = dualpa_max(norm_inf, lam_alpha)
-    f = norm_inf if norm_inf < lam_alpha else lam_alpha
-    c2 = lam_alpha/f
-
-    d_obj   = (2/n) * c2 * np.dot(R, y) - (c2**2) * (c1 * R_norm_sq + (lam_1_alpha/2)*b_norm_sq)
-    w_d_gap = (p_obj - d_obj)*n/np.linalg.norm(y)**2
-
-    return w_d_gap
-
-# @njit
-def dualpa_elnet(R, X, y, beta, lam_1_alpha, lam_alpha, ny2):
-    
-    # R_norm_sq = np.linalg.norm(R)**2
-    if lam_1_alpha == 0:
-        return dualpa(X, y, lam_alpha, beta, ny2)
-    
-    # R = y - X @ beta
-
-    n = len(y)
-    # b_norm_sq = np.linalg.norm(beta)**2
-    b_norm_sq = (beta**2).sum()
-
-    rt = (2/n) * R
-    norm_inf = np.linalg.norm(X.T @ rt  - lam_1_alpha*beta, ord=np.inf)
-
-    # f = dualpa_max(norm_inf, lam_alpha)
-    f = norm_inf if norm_inf < lam_alpha else lam_alpha
-    c2 = lam_alpha/f
-    c3 = 1 + c2**2
-
-    d_gap = (1/n)*np.dot(R, c3*R - 2*c2*y) +\
-            lam_alpha*np.linalg.norm(beta, ord=1) +\
-            c3*(lam_1_alpha/2)*b_norm_sq
-
-    return d_gap*n/ny2
-
-
 class Lasso:
     """
     lasso model.
@@ -241,12 +35,6 @@ class Lasso:
         self.seed = seed
 
         self.fit_intercept = fit_intercept
-        
-
-
-        # if self.warm_start:
-        #     self.beta = beta
-        # else:
         self.beta = np.array([])
 
 
@@ -261,9 +49,14 @@ class Lasso:
         self.Xj_T_y = np.array([])
         self.ny2 = np.array([])
 
-    # @njit
-    def soft_threshold(self, delta, lam, denom):
-        return msoft_threshold(delta, lam, denom)
+        self.y = np.array([])
+
+        self.x_offset = np.array([])
+        self.y_offset = 0.0
+
+        self.x_scale = np.array([])
+        self.y_scale = 1
+
         
     def initial_iterations(self, c1, Xj_T_y, Xj_T_Xj, X_T_X, all_p):
 
@@ -312,34 +105,39 @@ class Lasso:
             set_X (overwrite)
         """
 
-        if self.copyX and len(self.X) == 0:
-            self.X = copy.deepcopy(X)
-            self.X_T_X = np.matmul(X.T, X)
-            self.Xj_T_Xj = np.diag(self.X_T_X)
-            self.Xj_T_y = np.matmul(X.T, y)
-            self.ny2 = np.linalg.norm(y)**2
-
-        elif not self.copyX and len(self.X) == 0:
-            self.X = X
-            self.X_T_X = np.matmul(X.T, X)
-            self.Xj_T_Xj = np.diag(self.X_T_X)
-            self.Xj_T_y = np.matmul(X.T, y)
-            self.ny2 = np.linalg.norm(y)**2
-
-        elif self.copyX and len(self.X) != 0:
+        if self.copyX and len(self.X) != 0:
             pass
 
-        elif not self.copyX and len(self.X) != 0:
-            self.X = X
-            self.X_T_X = np.matmul(X.T, X)
+        else:
+            self.X = X.copy()
+            self.y = y.copy()
+            self.center_Xy()
+
+            # optimal for dot products in CD
+            self.X_T_X = np.asfortranarray( np.matmul(self.X.T, self.X) )
             self.Xj_T_Xj = np.diag(self.X_T_X)
-            self.Xj_T_y = np.matmul(X.T, y)
-            self.ny2 = np.linalg.norm(y)**2
+            self.Xj_T_y = np.matmul(self.X.T, self.y)
+            self.ny2 = np.dot(self.y, self.y)
+
+    def center_Xy(self):
+        """
+        center X and y
+        """
+        self.x_offset = np.mean(self.X, axis=0, dtype=self.X.dtype) # O(np)
+        self.x_scale  = np.std(self.X , axis=0, dtype=self.X.dtype) # O(np)
+        
+        zero_sd_indices = self.x_scale == 0 # O(p)
+        if np.any(zero_sd_indices): # O(p)
+            self.x_scale[zero_sd_indices] = 1 # O(p)
+
+        self.X = (self.X - self.x_offset) / self.x_scale # O(np)
+
+        self.y_offset = np.mean(self.y) # O(n)
+        self.y -= self.y_offset # O(n)
+
 
     def dual_gap(self, y):
         return dualpa(self.X, y, self.lam, self.beta, self.ny2)
-        # R = y - self.sparse_XB(self.X)
-        # return dualpa(R, self.X, y, self.lam, self.beta)
     
     def fit(self, X, y):
         # X = X_train
@@ -353,18 +151,15 @@ class Lasso:
 
         self.set_Xy(X,y)
 
-        # X_T_X = self.X_T_X
-        # # precompute Xj^T * Xj
-        # Xj_T_Xj = self.Xj_T_Xj
-        # # precompute Xj^T * y
-        # Xj_T_y = np.matmul(self.X.T, y)
         
         if not self.warm_start:
             # He-styled 
             # initialization 
             # of the coefficients
-            np.random.seed(self.seed)
-            self.beta = np.random.normal(0, np.sqrt(2/p), size=p) 
+            # np.random.seed(self.seed)
+            # self.beta = np.random.normal(0, np.sqrt(2/p), size=p) 
+
+            self.beta = np.zeros(p, dtype=self.X.dtype, order='F')  
          
         # few iterations of coordinate descent
         all_p = np.array(range(p))
@@ -422,7 +217,7 @@ class Lasso:
 
                     # R = y - self.sparse_XB(self.X)
                     # w_d_gap = dualpaR(R, self.X, y, self.lam, self.beta)
-                    w_d_gap = self.dual_gap(y)
+                    w_d_gap = self.dual_gap(self.y)
                     # w_d_gap = dualpa(self.X, y, self.lam, self.beta)
 
                     if w_d_gap < self.tol:
@@ -440,7 +235,8 @@ class Lasso:
             print("Model did not converge")
 
         if self.fit_intercept:
-            self.intercept = np.mean( y - self.sparse_XB(X) )
+            self.beta /= self.x_scale
+            self.intercept = self.y_offset - np.dot(self.x_offset, self.beta) # O(p)
             # print("Intercept: ", self.intercept)
 
     def cd_epoch(self, c1, chosen_ps, s_new, diff):
@@ -756,8 +552,7 @@ def k_fold_cv_random(X, y,
     return best_[0]
 
 
-def lasso_path(X_train, y_train, params, model, 
-               sequential_screening = True):
+def lasso_path(X_train, y_train, params, model, print_progress = True):
     """
     compute the lasso path based on the training set
     and  with errors based on the test set
@@ -782,78 +577,26 @@ def lasso_path(X_train, y_train, params, model,
     model.fit(X_train, y_train)
 
     path[:,0] = model.beta
+
+    if print_progress:
+        index_set = progressbar(range(1, len(lams)), "Computing lasso path: ", 40)
+        
+    else:
+        index_set = range(1, len(lams))
     
-    for i in range(1, len(lams)):
+    for i in index_set:
+    # for i in range(1, len(lams)):
 
         model.set_params(lam = lams[i],
                           warm_start = True, 
-                        #   prev_lam = lams[i-1],
                           prev_lam = None
                           )
             
-        model.fit(X_train, y_train)            
-        # errors[i] = None
+        model.fit(X_train, y_train)
         path[:,i] = model.beta
 
     return path
 
-def myenet_path(X_train, y_train, X_test, y_test, params, model, 
-               sequential_screening = True,
-               alpha = 0.5):
-    """
-    compute the lasso path based on the training set
-    and  with errors based on the test set
-    """
-    # model = Lasso()
-    # X = X_train
-    # y = y_train
-    # params = {'lam': np.logspace(-2, max_lambda(X,y), 3)}
-
-    if X_test is None and y_test is None:
-        X_test = X_train
-        y_test = y_train
-
-
-    _,p = X_train.shape
-    lams = params['lam']
-
-    errors = np.zeros(len(lams))
-    path = np.ones((p, len(params['lam'])))
-    y_train_enet = np.concatenate((y_train, np.zeros(p)))
-
-    all_gammas = np.zeros(len(lams))
-    for i,lam in enumerate(lams):
-        # i = 0
-        # lam = lams[i]
-        alfa = lam * alpha
-        beta = lam * (1 - alpha)
-        sf = (1 + beta) ** (-1/2)
-        gamma = alfa * sf
-        all_gammas[i] = gamma
-
-        if i > 1:
-            if sequential_screening:
-                prev_lam = all_gammas[i-1]
-
-            else:
-                prev_lam = None
-
-            model.set_params(lam = gamma,
-                             warm_start = True, 
-                             beta = model.beta,
-                             prev_lam = prev_lam)
-            
-        else:
-
-            model.set_params(lam=gamma)
-            # model.warm_start
-
-        X_train_enet =  np.vstack([X_train, np.sqrt(beta) * np.eye(p)])
-        model.fit(X_train_enet, y_train_enet)
-        errors[i] = None
-        path[:,i] = model.beta*sf
-
-    return path, all_gammas
 
 def theta(path_i):
 
@@ -919,5 +662,3 @@ def get_non_zero_coeffs(path, ZO, thresh = 0.5):
                 non_zero_coeffs.append(non_zero_idx)
 
     return non_zero_coeffs
-
-
